/*
    (C) 1995-96 AROS - The Amiga Research OS
    $Id: matchnext.c,v 1.2 2007/10/28 11:02:42 piru Exp $

    Desc:
    Lang: english
*/

#define	USE_SOFTLINKWORKAROUND	1

#include <exec/memory.h>
#include <exec/types.h>
#include <dos/dos.h>
#include <proto/exec.h>
#include "dos_intern.h"

#if USE_SOFTLINKWORKAROUND
static
LONG issoftlink(struct DosLibrary *DOSBase, CONST_STRPTR name);
static
BPTR makefakeslink(struct FileInfoBlock *fib, CONST_STRPTR name, LONG softlink);
#endif /* USE_SOFTLINKWORKAROUND */

/*****************************************************************************

    NAME */
#include <dos/dosasl.h>
#include <proto/dos.h>

	AROS_LH1(LONG, MatchNext,

/*  SYNOPSIS */
	AROS_LHA(struct AnchorPath *, AP, D1),

/*  LOCATION */
	struct DosLibrary *, DOSBase, 138, Dos)

/*  FUNCTION
	Find next file or directory that matches a given pattern.
	See <dos/dosasl.h> for more docs and how to control MatchNext().


    INPUTS
	AP  - pointer to Anchor Path structure which had been passed to
              MatchFirst() before.

    RESULT

    NOTES

    EXAMPLE

    BUGS

    SEE ALSO
	MatchFirst() MatchEnd() CurrentDir() Examine() ExNext()
	ParsePattern() <dos/dosasl.h>

    INTERNALS

    HISTORY
        09-11-2000  stegerg, rewritten from scratch.
	29-04-97    bergers, initial revision
	27-11-96    digulla automatically created from
			    dos_lib.fd and clib/dos_protos.h

*****************************************************************************/
{
    AROS_LIBFUNC_INIT
    AROS_LIBBASE_EXT_DECL(struct DosLibrary *,DOSBase)

    struct AChain 	*ac;
    BPTR 		origdir;
    LONG 		error = 0;
    UBYTE		dodir;

    DEBUG_MATCH(dprintf("MatchNext: anchor = $%lx\n", (ULONG) AP));

    FAILFORTASKNP(return ERROR_NOT_IMPLEMENTED;)

    if ((ULONG) AP & 3)
    {
	DEBUG_WARNINGS(ktprintf("MatchNext: BUG, AnchorPath 0x%lx isn't aligned by 4!\n", (ULONG) AP));

	/* Special implementation note: we really don't care about alignment! - Piru */
	//return ERROR_ ???;
    }

    ac = AP->ap_Current;

    origdir = CurrentDir(0);
    CurrentDir(origdir);

    //DEBUG_MATCH(dprintf("*** MatchNext: caller's pr_CurrentDir 0x%lx\n", (LONG) origdir));

    AP->ap_Flags &= (ULONG) ~(/*APF_DIDDIR |*/ APF_DirChanged);

    if (AP->ap_Flags & APF_NOMEMERR)
    {
	error = ERROR_NO_FREE_STORE;
	goto done;
    }

    if (!AP->ap_First)
    {
	error = ERROR_NO_MORE_ENTRIES;
	goto done;
    }

    dodir = AP->ap_Flags & APF_DODIR;
    AP->ap_Flags &= (ULONG) ~APF_DODIR;

    if (dodir &&
        !(AP->ap_Flags & APF_FollowHLinks) &&
        ac->an_Info.fib_DirEntryType == ST_LINKDIR)
    {
	AP->ap_Flags |= APF_DIDDIR;
    }
    else
    {

#if 1
	/* Cheap workaround for nonfilesystem calls (see MatchFirst) - Piru */
	if (ac->an_Flags & DDF_Single)
	{
	    ac->an_Flags &= ~DDF_Single;
	    error = ERROR_NO_MORE_ENTRIES;
	    goto done;
	}
#endif

	/*
	** Check if we are asked to enter a directory, but only do this
	** if it is really possible
	*/

	if (dodir &&
	    (ac->an_Flags & DDF_ExaminedBit) &&
	    (ac->an_Info.fib_DirEntryType > 0) &&
#if USE_SOFTLINKWORKAROUND
	    (ac->an_Info.fib_DirEntryType != ST_SOFTLINK ||
	     !(AP->ap_Extended & APEF_DontFollowSLinks)) &&
#endif /* USE_SOFTLINKWORKAROUND */
	    (ac->an_Child == NULL))
	{
	    DEBUG_MATCH(if (ac->an_Info.fib_DirEntryType == ST_SOFTLINK) dprintf("*** softlink, entry <%s> entrytype %ld\n", ac->an_Info.fib_FileName, ac->an_Info.fib_DirEntryType));
	    /*
	    ** Alloc a new AChain. Make it the active one. Set its string to "#?" and
	    ** mark it with DDF_AllBit Flag to indicate that this is a "APF_DODIR-AChain".
	    ** This is important for "back steppings", because "APF_DODIR-AChains" must
	    ** be removed and freed then and the user must be notified about the leaving
	    ** of a APF_DODIR-AChain with APF_DIDDIR.
	    */

	    if ((ac->an_Child = Match_AllocAChain(1, DOSBase)))
	    {
		ac->an_Child->an_Parent = ac;
		ac = ac->an_Child;
		AP->ap_Current = ac;

		ac->an_String[0] = P_ANY;
		ac->an_String[1] = 0;
		ac->an_Flags = DDF_PatternBit | DDF_AllBit;
		AP->ap_Flags |= APF_DirChanged;
	    }
	    else
	    {
		AP->ap_Flags |= APF_NOMEMERR;
		error = ERROR_NO_FREE_STORE;
		goto done;
	    }
	}


	/* Main loop for AChain traversing */

	for (;;)
	{
	    BOOL must_go_back = FALSE;

	    /* Check for user breaks (CTRL_C, ...) */

	    if (AP->ap_BreakBits)
	    {
		AP->ap_FoundBreak = CheckSignal(AP->ap_BreakBits);
		if (AP->ap_FoundBreak)
		{
		    error = ERROR_BREAK;
		    goto done;
		}
	    }

	    /* Check if AChain must be "setup"
	     */
	    if (!(ac->an_Flags & DDF_ExaminedBit))
	    {
		/*
		** This AChain must be "setup". First AChain->an_Lock must point
		** to the parent directory, that is the directory where this
		** AChain is "in". !
		*/

		if (ac->an_Parent)
		{
		    DEBUG_MATCH(dprintf("*** CurrentDir to parent lock 0x%lx\n", (LONG) ac->an_Parent->an_Lock));
		    CurrentDir(ac->an_Parent->an_Lock);
		    DEBUG_MATCH(dprintf("*** Lock \"%s\"\n", ac->an_Parent->an_Info.fib_FileName));
		    ac->an_Lock = Lock(ac->an_Parent->an_Info.fib_FileName, SHARED_LOCK);

		    if (!ac->an_Lock)
		    {
			error = IoErr();
			DEBUG_MATCH(dprintf("*** Lock() #1 ioerr: %ld\n", error));
			if (!error)
			    error = 500;
			goto done;
		    }
		}
#if !MATCHFUNCS_NO_DUPLOCK
		else
		{
		    ac->an_Lock = DupLock(origdir);
		    /* 07092002 bugfix: If zero lock, fail only if origdir was nonzero. - Piru
		     */
		    if (!ac->an_Lock && origdir)
		    {
			error = IoErr();
			DEBUG_MATCH(dprintf("*** DupLock() ioerr: %ld\n", error));
			if (!error)
			    error = 500;
			goto done;
		    }
		}
#else /* !MATCHFUNCS_NO_DUPLOCK */
		/*
		** If there was no ac->an_Parent then we are dealing with the
		** first AChain whose lock was already setup in Match_BuildAChainList
		*/
#endif /* !MATCHFUNCS_NO_DUPLOCK */

		DEBUG_MATCH(dprintf("*** CurrentDir to lock 0x%lx\n", (LONG) ac->an_Lock));
		CurrentDir(ac->an_Lock);

		if (ac->an_Flags & DDF_PatternBit)
		{
		    /*
		    ** If this is a pattern AChain we first Examine here our
		    ** parent directory, so that it then can be traversed with
		    ** ExNext
		    */

		    if (!Examine64(ac->an_Lock, &ac->an_Info, NULL))
		    {
			error = IoErr();
			DEBUG_MATCH(dprintf("*** Examine() #1 ioerr: %ld\n", error));
			if (!error)
			    error = 500;
			goto done;
		    }
		    ac->an_Flags |= DDF_ExaminedBit;

		    /*
		    ** If our "parent" is a file, we certainly can't go ExNexting
		    */

		    if (ac->an_Info.fib_DirEntryType < 0)
		    {
			/* [1] */

			must_go_back = TRUE;
		    }


		} /* if (ac->an_Flags & DDF_PatternBit) */
		else
		{
		    BPTR lock = 0;
		    LONG success;
#if USE_SOFTLINKWORKAROUND
		    LONG softlink = -2;
#endif /* USE_SOFTLINKWORKAROUND */

		    /*
		    ** This is a normal AChain (no pattern). Try to lock it
		    ** to see if it exists.
		    */

#if USE_SOFTLINKWORKAROUND

#warning "Experimental!!!"
		    if ((AP->ap_Extended & (APEF_LiteralSLinks | APEF_DontFollowSLinks)) &&
		        (softlink = issoftlink(DOSBase, ac->an_String)) >= 0)
		    {
			DEBUG_MATCH(dprintf("*** nopattern \"%s\" is a softlink, fake examine result\n", ac->an_String));

#warning "FIXME: Fix case with quoted patterns..."
			/* If literal softlinks are requested, use the name of the
			 * link instead of the link destination. IIRC this will give
			 * wrong case for certain 'quoted' patterns. - Piru
			 */

			if (1) //(AP->ap_Extended & APEF_LiteralSLinks)
			{
			    /* Fake Examine() result
			    */
			    lock = makefakeslink(&ac->an_Info, ac->an_String, softlink);
			}
		    }

		    if (!lock)
		    {
			lock = Lock(ac->an_String, SHARED_LOCK);
		    }
		    if (!lock)
		    {
			error = IoErr();
			DEBUG_MATCH(dprintf("*** MatchNext(), non-pattern, failed to Lock \"%s\" (ioerr %ld), try workaround\n", ac->an_String, (LONG) error));
			if (error == ERROR_OBJECT_NOT_FOUND)
			{
			    if (softlink >= -1 || (softlink = issoftlink(DOSBase, ac->an_String)) >= -1)
			    {
				if (softlink == -1)
				{
				    /* Drop into regular failure path
				    */
				    lock = 0;
				}
				else
				{
				    DEBUG_MATCH(dprintf("*** MatchNext() is softlink \"%s\", fake Examine result.\n", ac->an_String));

				    /* Fake Examine() result
				    */
				    lock = makefakeslink(&ac->an_Info, ac->an_String, softlink);
				}
			    }

			    SetIoErr(error);
			}
		    }

		    if (!lock)

#else /* USE_SOFTLINKWORKAROUND */

		    if (!(lock = Lock(ac->an_String, SHARED_LOCK)))

#endif /* USE_SOFTLINKWORKAROUND */

		    {
			DEBUG_MATCH(dprintf("*** Lock() #2 ioerr: %ld\n", IoErr()));

			/* It does not exist, so if possible go back one step
			 */
			/*DEBUG_MATCH(dprintf("*** parent 0x%lx, cur pat %ld parent pat %ld\n", (ULONG) ac->an_Parent,
			                    ac->an_Flags & DDF_PatternBit ? 1 : 0, ac->an_Parent && (ac->an_Parent->an_Flags & DDF_PatternBit) ? 1 : 0));
			*/
			//HSMOD: if (ac->an_Parent)
			//if (ac->an_Parent && (ac-an_Flags & DDF_PatternBit))
			if (ac->an_Parent && (ac->an_Parent->an_Flags & DDF_PatternBit))
			{
			    /* [2] */
			    //DEBUG_MATCH(dprintf("*** it has a parent & is a pattern, go back!\n"));

			    must_go_back = TRUE;
			}
			else
			{
			    /* if going back is not possible get error code and exit */

			    error = IoErr();
			    DEBUG_MATCH(dprintf("*** no parent, can't go back, Lock() #2 ioerr: %ld\n", error));
			    if (!error)
				error = 500;
			    goto done;
			}

		    } /* if (!(lock = Lock(ac->an_String, SHARED_LOCK))) */
		    else
		    {
			/* The File/Direcory ac->an_String exists
			 */
#if USE_SOFTLINKWORKAROUND
			if (lock == (BPTR) -1)
			{
			    /* It was a fake lock with faked Examine() result
			    */
			    success = DOSTRUE;
			}
			else
			{
#endif /* USE_SOFTLINKWORKAROUND */

			success = Examine64(lock, &ac->an_Info, NULL);
			UnLock(lock);

#if USE_SOFTLINKWORKAROUND
			}
#endif /* USE_SOFTLINKWORKAROUND */

			if (!success)
			{
			    /*
			    ** Examine()ing the file/directory did not
			    ** work, although the lock was successful!?.
			    ** Get error code and exit
			    */

			    error = IoErr();
			    DEBUG_MATCH(dprintf("*** Examine() #2 ioerr: %ld\n", error));
			    if (!error)
				error = 500;
			    goto done;
			}

			/*
			** This strcpy is necessary, because in case
			** of empty ac->an_String("") fib_FileName would
			** get parent directory name which it must not!
			*/
#if 1
			/*
			** Only replace the name if it's empty or path.
			** fixes the problem where ticked patterns got
			** turned into uppercase ("test'#" -> TEST#).
			** - Piru
			*/
			if (*ac->an_String == '\0' || FilePart(ac->an_String) != ac->an_String)
			{
			    strcpy(ac->an_Info.fib_FileName, ac->an_String);
			}
#else
			strcpy(ac->an_Info.fib_FileName, ac->an_String);
#endif

			ac->an_Flags |= DDF_ExaminedBit;

			/*
			** If this is a file, but there are still more path components to
			** follow then we have to go back one step (AChain)
			*/

			if (ac->an_Child && (ac->an_Info.fib_DirEntryType < 0))
			{
			    /* [1] */
			    //DEBUG_MATCH(dprintf("*** must go back, entry <%s> entrytype %ld\n", ac->an_Info.fib_FileName, ac->an_Info.fib_DirEntryType));

			    must_go_back = TRUE;
			}

			/*
			** Here we either have found a matching file/directory (result)
			** or, if ac->an_Child != NULL we have still to continue walking
			** through the AChains until we are in the last one. This all
			** happens further below
			*/

		    } /* if (!(lock = Lock(ac->an_String, SHARED_LOCK))) else ... */

		} /* if (ac->an_Flags & DDF_PatternBit) else ... */

	    } /* if (!(ac->an_Flags & DDF_ExaminedBit)) */
	    else
	    {
		/*
		** This AChain was already setup.
		**
		** When an AChain which is *not* a pattern already had DDF_PatternBit
		** set, then this means ERROR_NO_MORE_ENTRIES, so we try to go back
		** one step
		*/

		if (!(ac->an_Flags & DDF_PatternBit))
		{
		    /* [4] */

		    must_go_back = TRUE;
		}
	    }

	    /*
	    ** Here we can be sure that the actual AChain is setup, ie: it will
	    ** have ac->an_Lock set correctly and to indicate this DDF_ExaminedBit
	    ** was set
	    */

	    if (!must_go_back)
	    {
		//DEBUG_MATCH(dprintf("*** !must_go_back, CurrentDir to lock 0x%lx\n", (LONG) ac->an_Lock));
		CurrentDir(ac->an_Lock);

		if (ac->an_Flags & DDF_PatternBit)
		{
		    if (ExNext64(ac->an_Lock, &ac->an_Info, NULL))
		    {
			if (MatchPatternNoCase(ac->an_String, ac->an_Info.fib_FileName))
			{
			    /*
			    ** This file matches the pattern in ac->an_String. If there
			    ** are no more AChains to follow then we have found a matching
			    ** file/directory (a result)  --> break.
			    */
			    if (!ac->an_Child)
				break;

#if 1
			    /* There is a AChain entry below us, it mean the current chain
			    ** entry must be a directory to match. - Piru
			    */
			    if (ac->an_Info.fib_DirEntryType < 0)
			    {
				/* Was a file so didn't really match. Go to top of
				** "for (;;)" loop.
				*/
				continue;
			    }

			    //DEBUG_MATCH(dprintf("*** matched entry <%s> entrytype %ld\n", ac->an_Info.fib_FileName, ac->an_Info.fib_DirEntryType));
#endif
			}
			else
			{
			    /* Did not match. Go to top of "for (;;)" loop
			     */
			    continue;
			}
		    }
		    else
		    {
			error = IoErr();
			DEBUG_MATCH(dprintf("*** ExNext() #1 ioerr: %ld\n", error));
			if (!error)
			    error = 500;
			if (error != ERROR_NO_MORE_ENTRIES)
			    goto done;

			/* [3] */

			must_go_back = TRUE;
		    }
		} /* if (ac->an_Flags & DDF_PatternBit) */
		/*else
		{
		    DEBUG_MATCH(dprintf("*** no pattern, matched entry <%s> entrytype %ld\n", ac->an_Info.fib_FileName, ac->an_Info.fib_DirEntryType));
		}*/

	    } /* if (!must_go_back) */

	    /*
	    ** Handle the cases where we must (try to) go back to the previous AChain.
	    ** This can happen if the actual AChain turned out to be a file although
	    ** there are still more AChains to follow [1]. Or if the actual AChain did not
	    ** exist at all [2]. Or if in a pattern AChain ExNext() told us that there are
	    ** no more entries [3]. Or if we were getting to a normal (no pattern) AChain
	    ** which was already setup (DDF_ExaminedBit) [4].
	    */

	    if (must_go_back)
	    {
		/* Check if going back is possible at all */

		if (!ac->an_Parent)
		{
		    error = ERROR_NO_MORE_ENTRIES;
		    goto done;
		}

		/* Yep. It is possible. So let's cleanup the AChain.
		 */
		//DEBUG_MATCH(dprintf("*** must_go_back, CurrentDir to parent lock 0x%lx\n", (LONG) ac->an_Parent->an_Lock));
		CurrentDir(ac->an_Parent->an_Lock);

		UnLock(ac->an_Lock);

		ac->an_Lock = NULL;
		ac->an_Flags &= (ULONG) ~DDF_ExaminedBit;

		/* Make ac and AP->ap_Current point to the previous AChain
		 */
		AP->ap_Current = ac->an_Parent;
		AP->ap_Flags |= APF_DirChanged;

		/*
		** If this was an APF_DODIR Achain (indicated by DDF_AllBit)
		** then the AChain must be unlinked and freed. And the user
		** must be informed about the leaving with APF_DIDDIR and
		** a "result" in AnchorPath which points to the directory which
		** was left.
		*/

		if (ac->an_Flags & DDF_AllBit)
		{
		    AP->ap_Current->an_Child = NULL;
		    Match_FreeAChain(ac, DOSBase);
		    AP->ap_Flags |= APF_DIDDIR;

		    /* go out of for (;;) loop --> MakeResult
		     */
		    break;
		}

		ac = AP->ap_Current;

	    } /* if (must_go_back) */
	    else
	    {
		if (!ac->an_Child)
		{
		    /*
		    ** We have reached the last AChain. And this means that
		    ** we have found a matching file/directory :-)). Go out of
		    ** for (;;) loop --> MakeResult
		    */

		    break;
		}

		ac = ac->an_Child;
		AP->ap_Current = ac;
		AP->ap_Flags |= APF_DirChanged;
	    }

	} /* for (;;) */
    }

    error = Match_MakeResult(AP, DOSBase);

done:
    CurrentDir(origdir);

    /* If error occured, free resources and end match
     */
    if (error)
	MatchEnd(AP);

    SetIoErr(error);

    DEBUG_MATCH(dprintf("MatchNext: err = %ld, name = \"%s\"\n", error, error ? "" : AP->ap_Info.fib_FileName));

    return error;

    AROS_LIBFUNC_EXIT

} /* MatchNext */


#if USE_SOFTLINKWORKAROUND

static
LONG issoftlink(struct DosLibrary *DOSBase, CONST_STRPTR name)
{
    struct DevProc *dvp;
    LONG ret = -2; /* default: error */

    //DEBUG_MATCH(dprintf("*** issoftlink entry: \"%s\"\n", name, ret));

#warning "Add multiassign support?"
    dvp = GetDeviceProc(name, NULL);
    if (dvp)
    {
	UBYTE *tmp;

	tmp = dosAllocMem(256, MEMF_PUBLIC);
	if (tmp)
	{
	    BPTR lock;

	    MakeRelBSTR(tmp, name);

	    lock = DoPkt3(dvp->dvp_Port, ACTION_LOCATE_OBJECT,
	                  dvp->dvp_Lock, MKBADDR(tmp), SHARED_LOCK);
	    if (lock)
	    {
		struct FileLock *fl = BADDR(lock);

		if (fl->fl_Task)
		{
		    struct MsgPort *port = fl->fl_Task;
		    fl->fl_Task = NULL;
		    DoPkt1(port, ACTION_FREE_LOCK, lock);
		}

		/* not a softlink */
		ret = -1;
	    }
	    else if (IoErr() == ERROR_IS_SOFT_LINK &&
	             ReadLink(dvp->dvp_Port, dvp->dvp_Lock,
	                      name, tmp, 256 - 1) >= 0)
	    {
		/* is a softlink! */
		tmp[255] = '\0';
		ret = strlen(tmp);
	    }

	    dosFreeVec(tmp);
	}

	FreeDeviceProc(dvp);
    }

    DEBUG_MATCH(dprintf("*** issoftlink: \"%s\" ret %ld\n", name, ret));
    return ret;
}

static
BPTR makefakeslink(struct FileInfoBlock *fib, CONST_STRPTR name, LONG softlink)
{
    fib->fib_DiskKey            = 0;
    fib->fib_DirEntryType       = ST_SOFTLINK;
    strcpy(fib->fib_FileName, name);
    fib->fib_Protection         = 0;
    fib->fib_EntryType          = ST_SOFTLINK;
    fib->fib_Size               = softlink;
    fib->fib_NumBlocks          = 1;
    fib->fib_Date.ds_Days       = 0;
    fib->fib_Date.ds_Minute     = 0;
    fib->fib_Date.ds_Tick       = 0;
    fib->fib_Comment[0]         = '\0';
    fib->fib_OwnerUID           = 0;
    fib->fib_OwnerGID           = 0;
    fib->fib_Size64             = softlink;
    fib->fib_NumBlocks64        = 1;

    /* Fake "lock"
    */
    return -1;
}

#endif /* USE_SOFTLINKWORKAROUND */

